@@ -70,7 +70,7 @@ class LsDelCommand(cmd.BaseCommand):
             raise RuntimeError(msg)
 
 
-class LsListCommand(cmd.BaseCommand):
+class LsListCommand(cmd.ReadOnlyCommand):
     def run_idl(self, txn):
         table = self.api.tables['Logical_Switch']
         self.result = [rowview.RowView(r) for r in table.rows.values()]
@@ -190,7 +190,7 @@ class PgAclDelCommand(_AclDelHelper):
     lookup_table = 'Port_Group'
 
 
-class _AclListHelper(cmd.BaseCommand):
+class _AclListHelper(cmd.ReadOnlyCommand):
     def __init__(self, api, entity):
         super(_AclListHelper, self).__init__(api)
         self.entity = entity
@@ -247,10 +247,9 @@ class QoSAddCommand(cmd.AddCommand):
 
     def run_idl(self, txn):
         ls = self.api.lookup('Logical_Switch', self.switch)
-        qos_rules = [row for row in ls.qos_rules if self.qos_match(row)]
-        if qos_rules:
+        for qos_rule in iter(r for r in ls.qos_rules if self.qos_match(r)):
             if self.may_exist:
-                self.result = rowview.RowView(qos_rules[0])
+                self.result = rowview.RowView(qos_rule)
                 return
             raise RuntimeError("QoS (%s, %s, %s) already exists" % (
                 self.direction, self.priority, self.match))
@@ -271,7 +270,7 @@ class QoSAddCommand(cmd.AddCommand):
 
 class QoSDelCommand(cmd.BaseCommand):
     def __init__(self, api, switch, direction=None,
-                 priority=None, match=None):
+                 priority=None, match=None, if_exists=True):
         if (priority is None) != (match is None):
             raise TypeError("Must specify priority and match together")
         if priority is not None and not direction:
@@ -279,6 +278,7 @@ class QoSDelCommand(cmd.BaseCommand):
         super(QoSDelCommand, self).__init__(api)
         self.switch = switch
         self.conditions = []
+        self.if_exists = if_exists
         if direction:
             self.conditions.append(('direction', '=', direction))
             # priority can be 0
@@ -287,14 +287,21 @@ class QoSDelCommand(cmd.BaseCommand):
                                     ('match', '=', match)]
 
     def run_idl(self, txn):
-        ls = self.api.lookup('Logical_Switch', self.switch)
+        try:
+            ls = self.api.lookup('Logical_Switch', self.switch)
+        except idlutils.RowNotFound:
+            if self.if_exists:
+                return
+            msg = 'Logical Switch %s does not exist' % self.switch
+            raise RuntimeError(msg)
+
         for row in ls.qos_rules:
             if idlutils.row_match(row, self.conditions):
                 ls.delvalue('qos_rules', row)
                 row.delete()
 
 
-class QoSListCommand(cmd.BaseCommand):
+class QoSListCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, switch):
         super(QoSListCommand, self).__init__(api)
         self.switch = switch
@@ -304,6 +311,31 @@ class QoSListCommand(cmd.BaseCommand):
         self.result = [rowview.RowView(row) for row in ls.qos_rules]
 
 
+class QoSDelExtIdCommand(cmd.BaseCommand):
+    def __init__(self, api, lswitch, external_ids, if_exists=False):
+        if not external_ids:
+            raise TypeError('external_ids dictionary cannot be empty')
+        super(QoSDelExtIdCommand, self).__init__(api)
+        self.lswitch = lswitch
+        self.external_ids = external_ids
+        self.if_exists = if_exists
+
+    def run_idl(self, txn):
+        try:
+            lswitch = idlutils.row_by_value(self.api.idl, 'Logical_Switch',
+                                            'name', self.lswitch)
+        except idlutils.RowNotFound:
+            if self.if_exists:
+                return
+            msg = 'Logical Switch %s does not exist' % self.lswitch
+            raise RuntimeError(msg)
+
+        for qos in lswitch.qos_rules:
+            if self.external_ids.items() <= qos.external_ids.items():
+                lswitch.delvalue('qos_rules', qos)
+                qos.delete()
+
+
 class LspAddCommand(cmd.AddCommand):
     table_name = 'Logical_Switch_Port'
 
@@ -398,7 +430,7 @@ class LspDelCommand(PortDelCommand):
             if_exists)
 
 
-class LspListCommand(cmd.BaseCommand):
+class LspListCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, switch=None):
         super(LspListCommand, self).__init__(api)
         self.switch = switch
@@ -415,7 +447,7 @@ class LspGetCommand(cmd.BaseGetRowCommand):
     table = 'Logical_Switch_Port'
 
 
-class LspGetParentCommand(cmd.BaseCommand):
+class LspGetParentCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, port):
         super(LspGetParentCommand, self).__init__(api)
         self.port = port
@@ -425,7 +457,7 @@ class LspGetParentCommand(cmd.BaseCommand):
         self.result = next(iter(lsp.parent_name), "")
 
 
-class LspGetTagCommand(cmd.BaseCommand):
+class LspGetTagCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, port):
         super(LspGetTagCommand, self).__init__(api)
         self.port = port
@@ -443,8 +475,8 @@ class LspSetAddressesCommand(cmd.BaseCommand):
         for addr in addresses:
             if not self.addr_re.match(addr):
                 raise TypeError(
-                    "address must be router/unknown/dynamic/"
-                    "ethaddr[ ipaddr...]")
+                    "address (%s) must be router/unknown/dynamic/"
+                    "ethaddr[ ipaddr...]" % (addr,))
         super(LspSetAddressesCommand, self).__init__(api)
         self.port = port
         self.addresses = addresses
@@ -454,7 +486,7 @@ class LspSetAddressesCommand(cmd.BaseCommand):
         lsp.addresses = self.addresses
 
 
-class LspGetAddressesCommand(cmd.BaseCommand):
+class LspGetAddressesCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, port):
         super(LspGetAddressesCommand, self).__init__(api)
         self.port = port
@@ -477,7 +509,7 @@ class LspSetPortSecurityCommand(cmd.BaseCommand):
         lsp.port_security = self.addresses
 
 
-class LspGetPortSecurityCommand(cmd.BaseCommand):
+class LspGetPortSecurityCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, port):
         super(LspGetPortSecurityCommand, self).__init__(api)
         self.port = port
@@ -487,7 +519,7 @@ class LspGetPortSecurityCommand(cmd.BaseCommand):
         self.result = lsp.port_security
 
 
-class LspGetUpCommand(cmd.BaseCommand):
+class LspGetUpCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, port):
         super(LspGetUpCommand, self).__init__(api)
         self.port = port
@@ -509,7 +541,7 @@ class LspSetEnabledCommand(cmd.BaseCommand):
         lsp.enabled = self.is_enabled
 
 
-class LspGetEnabledCommand(cmd.BaseCommand):
+class LspGetEnabledCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, port):
         super(LspGetEnabledCommand, self).__init__(api)
         self.port = port
@@ -531,7 +563,7 @@ class LspSetTypeCommand(cmd.BaseCommand):
         lsp.type = self.port_type
 
 
-class LspGetTypeCommand(cmd.BaseCommand):
+class LspGetTypeCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, port):
         super(LspGetTypeCommand, self).__init__(api)
         self.port = port
@@ -542,23 +574,27 @@ class LspGetTypeCommand(cmd.BaseCommand):
 
 
 class LspSetOptionsCommand(cmd.BaseCommand):
+    table = 'Logical_Switch_Port'
+
     def __init__(self, api, port, **options):
         super(LspSetOptionsCommand, self).__init__(api)
         self.port = port
         self.options = options
 
     def run_idl(self, txn):
-        lsp = self.api.lookup('Logical_Switch_Port', self.port)
+        lsp = self.api.lookup(self.table, self.port)
         lsp.options = self.options
 
 
-class LspGetOptionsCommand(cmd.BaseCommand):
+class LspGetOptionsCommand(cmd.ReadOnlyCommand):
+    table = 'Logical_Switch_Port'
+
     def __init__(self, api, port):
         super(LspGetOptionsCommand, self).__init__(api)
         self.port = port
 
     def run_idl(self, txn):
-        lsp = self.api.lookup('Logical_Switch_Port', self.port)
+        lsp = self.api.lookup(self.table, self.port)
         self.result = lsp.options
 
 
@@ -573,7 +609,7 @@ class LspSetDhcpV4OptionsCommand(cmd.BaseCommand):
         lsp.dhcpv4_options = self.dhcpopt_uuid
 
 
-class LspGetDhcpV4OptionsCommand(cmd.BaseCommand):
+class LspGetDhcpV4OptionsCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, port):
         super(LspGetDhcpV4OptionsCommand, self).__init__(api)
         self.port = port
@@ -610,7 +646,7 @@ class DhcpOptionsDelCommand(cmd.BaseCommand):
         dhcpopt.delete()
 
 
-class DhcpOptionsListCommand(cmd.BaseCommand):
+class DhcpOptionsListCommand(cmd.ReadOnlyCommand):
     def run_idl(self, txn):
         self.result = [rowview.RowView(r) for
                        r in self.api.tables['DHCP_Options'].rows.values()]
@@ -631,7 +667,7 @@ class DhcpOptionsSetOptionsCommand(cmd.BaseCommand):
         dhcpopt.options = self.options
 
 
-class DhcpOptionsGetOptionsCommand(cmd.BaseCommand):
+class DhcpOptionsGetOptionsCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, dhcpopt_uuid):
         super(DhcpOptionsGetOptionsCommand, self).__init__(api)
         self.dhcpopt_uuid = dhcpopt_uuid
@@ -686,7 +722,7 @@ class LrDelCommand(cmd.BaseCommand):
             raise RuntimeError(msg)
 
 
-class LrListCommand(cmd.BaseCommand):
+class LrListCommand(cmd.ReadOnlyCommand):
     def run_idl(self, txn):
         self.result = [rowview.RowView(r) for
                        r in self.api.tables['Logical_Router'].rows.values()]
@@ -723,7 +759,7 @@ class LrpAddCommand(cmd.BaseCommand):
                     msg = "Port %s exists with different networks" % (
                         self.port)
                 elif (not self.peer) != (not lrp.peer) or (
-                    self.peer != lrp.peer):
+                        self.peer != lrp.peer):
                     msg = "Port %s exists with different peer" % (self.port)
                 if msg:
                     raise RuntimeError(msg)
@@ -739,6 +775,13 @@ class LrpAddCommand(cmd.BaseCommand):
         if self.peer:
             lrp.peer = self.peer
         lr.addvalue('ports', lrp)
+        gwcs = self.columns.pop('gateway_chassis', [])
+        for n, chassis in enumerate(gwcs):
+            gwc_name = '%s_%s' % (lrp.name, chassis)
+            cmd = GatewayChassisAddCommand(self.api, gwc_name, chassis,
+                                           len(gwcs) - n, may_exist=True)
+            cmd.run_idl(txn)
+            lrp.addvalue('gateway_chassis', cmd.result)
         self.set_columns(lrp, **self.columns)
         self.result = lrp.uuid
 
@@ -756,7 +799,7 @@ class LrpDelCommand(PortDelCommand):
             if_exists)
 
 
-class LrpListCommand(cmd.BaseCommand):
+class LrpListCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, router):
         super(LrpListCommand, self).__init__(api)
         self.router = router
@@ -777,7 +820,7 @@ class LrpSetEnabledCommand(cmd.BaseCommand):
         lrp.enabled = self.is_enabled
 
 
-class LrpGetEnabledCommand(cmd.BaseCommand):
+class LrpGetEnabledCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, port):
         super(LrpGetEnabledCommand, self).__init__(api)
         self.port = port
@@ -788,6 +831,14 @@ class LrpGetEnabledCommand(cmd.BaseCommand):
         self.result = next(iter(lrp.enabled), True)
 
 
+class LrpSetOptionsCommand(LspSetOptionsCommand):
+    table = 'Logical_Router_Port'
+
+
+class LrpGetOptionsCommand(LspGetOptionsCommand):
+    table = 'Logical_Router_Port'
+
+
 class LrRouteAddCommand(cmd.BaseCommand):
     def __init__(self, api, router, prefix, nexthop, port=None,
                  policy='dst-ip', may_exist=False):
@@ -858,7 +909,7 @@ class LrRouteDelCommand(cmd.BaseCommand):
             raise RuntimeError(msg)
 
 
-class LrRouteListCommand(cmd.BaseCommand):
+class LrRouteListCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, router):
         super(LrRouteListCommand, self).__init__(api)
         self.router = router
@@ -941,7 +992,14 @@ class LrNatDelCommand(cmd.BaseCommand):
                 raise TypeError("nat_type not in %s" % str(const.NAT_TYPES))
             self.conditions += [('type', '=', nat_type)]
             if match_ip:
-                match_ip = str(netaddr.IPAddress(match_ip))
+                try:
+                    match_ip = str(netaddr.IPAddress(match_ip))
+                except ValueError:
+                    # logical_ip can be IPNetwork
+                    if nat_type == const.NAT_SNAT:
+                        match_ip = str(netaddr.IPNetwork(match_ip))
+                    else:
+                        raise
                 self.col = ('logical_ip' if nat_type == const.NAT_SNAT
                             else 'external_ip')
                 self.conditions += [(self.col, '=', match_ip)]
@@ -967,7 +1025,7 @@ class LrNatDelCommand(cmd.BaseCommand):
                                        match=self.match_ip)
 
 
-class LrNatListCommand(cmd.BaseCommand):
+class LrNatListCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, router):
         super(LrNatListCommand, self).__init__(api)
         self.router = router
@@ -1022,19 +1080,22 @@ class LbDelCommand(cmd.BaseCommand):
     def run_idl(self, txn):
         try:
             lb = self.api.lookup('Load_Balancer', self.lb)
-            if self.vip:
-                if self.vip in lb.vips:
-                    if self.if_exists:
-                        return
-                    lb.delkey('vips', self.vip)
-            else:
-                lb.delete()
         except idlutils.RowNotFound:
             if not self.if_exists:
                 raise
+            return
+        if self.vip:
+            if self.vip in lb.vips:
+                lb.delkey('vips', self.vip)
+            elif not self.if_exists:
+                raise idlutils.RowNotFound(table='Load_Balancer', col=self.vip,
+                                           match=self.lb)
+        # Remove load balancer if vips were not provided or no vips are left.
+        if not self.vip or not lb.vips:
+            lb.delete()
 
 
-class LbListCommand(cmd.BaseCommand):
+class LbListCommand(cmd.ReadOnlyCommand):
     def run_idl(self, txn):
         self.result = [rowview.RowView(r)
                        for r in self.api.tables['Load_Balancer'].rows.values()]
@@ -1079,7 +1140,7 @@ class LrLbDelCommand(cmd.BaseCommand):
             raise
 
 
-class LrLbListCommand(cmd.BaseCommand):
+class LrLbListCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, router):
         super(LrLbListCommand, self).__init__(api)
         self.router = router
@@ -1128,7 +1189,7 @@ class LsLbDelCommand(cmd.BaseCommand):
             raise
 
 
-class LsLbListCommand(cmd.BaseCommand):
+class LsLbListCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, switch):
         super(LsLbListCommand, self).__init__(api)
         self.switch = switch
@@ -1163,7 +1224,7 @@ class DnsGetCommand(cmd.BaseGetRowCommand):
     table = 'DNS'
 
 
-class DnsListCommand(cmd.BaseCommand):
+class DnsListCommand(cmd.ReadOnlyCommand):
     def run_idl(self, txn):
         table = self.api.tables['DNS']
         self.result = [rowview.RowView(r) for r in table.rows.values()]
@@ -1291,3 +1352,138 @@ class PgDelPortCommand(_PgUpdatePortsHelper):
 
 class PgGetCommand(cmd.BaseGetRowCommand):
     table = 'Port_Group'
+
+
+class GatewayChassisAddCommand(cmd.AddCommand):
+    table_name = 'Gateway_Chassis'
+
+    def __init__(self, api, name, chassis_name, priority=0, may_exist=False,
+                 **columns):
+        super(GatewayChassisAddCommand, self).__init__(api)
+        self.name = name
+        self.chassis_name = chassis_name
+        self.priority = priority
+        self.may_exist = may_exist
+        self.columns = columns
+
+    def run_idl(self, txn):
+        if self.may_exist:
+            gwc = self.api.lookup(self.table_name, self.name, None)
+        else:
+            gwc = None
+        if not gwc:
+            # If gwc exists with name, this will properly fail if not may_exist
+            # since 'name' is indexed
+            gwc = txn.insert(self.api.tables[self.table_name])
+            gwc.name = self.name
+        gwc.priority = self.priority
+        self.set_columns(gwc, **self.columns)
+        self.result = gwc
+
+
+class HAChassisGroupAddCommand(cmd.AddCommand):
+    table_name = 'HA_Chassis_Group'
+
+    def __init__(self, api, name, may_exist=False, **columns):
+        super(HAChassisGroupAddCommand, self).__init__(api)
+        self.name = name
+        self.may_exist = may_exist
+        self.columns = columns
+
+    def run_idl(self, txn):
+        if self.may_exist:
+            try:
+                hcg = self.api.lookup(self.table_name, self.name)
+                self.result = rowview.RowView(hcg)
+                return
+            except idlutils.RowNotFound:
+                pass
+
+        hcg = txn.insert(self.api._tables[self.table_name])
+        hcg.name = self.name
+        self.set_columns(hcg, **self.columns)
+        self.result = hcg.uuid
+
+
+class HAChassisGroupDelCommand(cmd.BaseCommand):
+    table_name = 'HA_Chassis_Group'
+
+    def __init__(self, api, name, if_exists=False):
+        super(HAChassisGroupDelCommand, self).__init__(api)
+        self.name = name
+        self.if_exists = if_exists
+
+    def run_idl(self, txn):
+        try:
+            hcg = self.api.lookup(self.table_name, self.name)
+            hcg.delete()
+        except idlutils.RowNotFound:
+            if self.if_exists:
+                return
+            raise RuntimeError(
+                'HA Chassis Group %s does not exist' % self.name)
+
+
+class HAChassisGroupGetCommand(cmd.BaseGetRowCommand):
+    table = 'HA_Chassis_Group'
+
+
+class HAChassisGroupAddChassisCommand(cmd.AddCommand):
+    table_name = 'HA_Chassis'
+
+    def __init__(self, api, hcg_id, chassis, priority, **columns):
+        super(HAChassisGroupAddChassisCommand, self).__init__(api)
+        self.hcg_id = hcg_id
+        self.chassis = chassis
+        self.priority = priority
+        self.columns = columns
+
+    def run_idl(self, txn):
+        hc_group = self.api.lookup('HA_Chassis_Group', self.hcg_id)
+        found = False
+        hc = None
+        for hc in hc_group.ha_chassis:
+            if hc.chassis_name != self.chassis:
+                continue
+            found = True
+            break
+        else:
+            hc = txn.insert(self.api.tables[self.table_name])
+            hc.chassis_name = self.chassis
+
+        hc.priority = self.priority
+        self.set_columns(hc, **self.columns)
+        if not found:
+            hc_group.addvalue('ha_chassis', hc)
+
+        self.result = hc.uuid
+
+
+class HAChassisGroupDelChassisCommand(cmd.BaseCommand):
+    table_name = 'HA_Chassis'
+
+    def __init__(self, api, hcg_id, chassis, if_exists=False):
+        super(HAChassisGroupDelChassisCommand, self).__init__(api)
+        self.hcg_id = hcg_id
+        self.chassis = chassis
+        self.if_exists = if_exists
+
+    def run_idl(self, txn):
+        try:
+            hc_group = self.api.lookup('HA_Chassis_Group', self.hcg_id)
+        except idlutils.RowNotFound:
+            if self.if_exists:
+                return
+
+        hc = None
+        for hc in hc_group.ha_chassis:
+            if hc.chassis_name == self.chassis:
+                break
+        else:
+            if self.if_exists:
+                return
+            raise RuntimeError(
+                'HA Chassis %s does not exist' % self.hcg_id)
+
+        hc_group.delvalue('ha_chassis', hc)
+        hc.delete()
