@@ -36,7 +36,8 @@ class TestLogicalSwitchOps(OvnNorthboundTest):
         self.table = self.api.tables['Logical_Switch']
 
     def _ls_add(self, *args, **kwargs):
-        fix = self.useFixture(fixtures.LogicalSwitchFixture(*args, **kwargs))
+        fix = self.useFixture(fixtures.LogicalSwitchFixture(self.api, *args,
+                                                            **kwargs))
         self.assertIn(fix.obj.uuid, self.table.rows)
         return fix.obj
 
@@ -97,8 +98,7 @@ class TestLogicalSwitchOps(OvnNorthboundTest):
         self.api.ls_del(name, if_exists=True).execute(check_error=True)
 
     def test_ls_list(self):
-        with self.api.transaction(check_error=True):
-            switches = {self._ls_add() for _ in range(3)}
+        switches = {self._ls_add() for _ in range(3)}
         switch_set = set(self.api.ls_list().execute(check_error=True))
         self.assertTrue(switches.issubset(switch_set))
 
@@ -106,8 +106,10 @@ class TestLogicalSwitchOps(OvnNorthboundTest):
 class TestAclOps(OvnNorthboundTest):
     def setUp(self):
         super(TestAclOps, self).setUp()
-        self.switch = self.useFixture(fixtures.LogicalSwitchFixture()).obj
-        self.port_group = self.useFixture(fixtures.PortGroupFixture()).obj
+        self.switch = self.useFixture(
+            fixtures.LogicalSwitchFixture(self.api)).obj
+        self.port_group = self.useFixture(
+            fixtures.PortGroupFixture(self.api)).obj
 
     def _acl_add(self, entity, *args, **kwargs):
         self.assertIn(entity, ['lswitch', 'port_group'])
@@ -219,7 +221,8 @@ class TestAclOps(OvnNorthboundTest):
 class TestQoSOps(OvnNorthboundTest):
     def setUp(self):
         super(TestQoSOps, self).setUp()
-        self.switch = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+        self.switch = self.useFixture(
+            fixtures.LogicalSwitchFixture(self.api)).obj
 
     def _qos_add(self, *args, **kwargs):
         cmd = self.api.qos_add(self.switch.uuid, *args, **kwargs)
@@ -307,6 +310,13 @@ class TestQoSOps(OvnNorthboundTest):
         self.assertRaises(TypeError, self.api.qos_del, self.switch.uuid,
                           priority=0)
 
+    def test_qos_del_ls_not_present_if_exists_true(self):
+        self.api.qos_del('some_other_ls').execute(check_error=True)
+
+    def test_qos_del_ls_not_present_if_exists_false(self):
+        cmd = self.api.qos_del('some_other_ls', if_exists=False)
+        self.assertRaises(RuntimeError, cmd.execute, check_error=True)
+
     def test_qos_list(self):
         r1 = self._qos_add('from-lport', 0, 'output == "fake_port"', 1000)
         r2 = self._qos_add('from-lport', 1, 'output == "fake_port2"', 1000)
@@ -315,13 +325,76 @@ class TestQoSOps(OvnNorthboundTest):
         self.assertIn(r1, qos_rules)
         self.assertIn(r2, qos_rules)
 
+    def _create_fip_qoses(self):
+        ext_ids_1 = {'key1': 'value1', 'key2': 'value2'}
+        self.qos_1 = self._qos_add('from-lport', 0, 'output == "fake_port1"',
+                                   dscp=11, external_ids=ext_ids_1)
+        ext_ids_2 = {'key3': 'value3', 'key4': 'value4'}
+        self.qos_2 = self._qos_add('from-lport', 1, 'output == "fake_port2"',
+                                   dscp=11, external_ids=ext_ids_2)
+        self.qos_3 = self._qos_add('from-lport', 2, 'output == "fake_port3"',
+                                   dscp=10)
+
+    def test_qos_delete_external_ids(self):
+        self._create_fip_qoses()
+        self.api.qos_del_ext_ids(self.switch.name,
+                                 {'key1': 'value1'}).execute(check_error=True)
+        qos_rules = self.api.qos_list(
+            self.switch.uuid).execute(check_error=True)
+        self.assertCountEqual([self.qos_2, self.qos_3], qos_rules)
+
+        self.api.qos_del_ext_ids(
+            self.switch.name,
+            {'key3': 'value3', 'key4': 'value4'}).execute(check_error=True)
+        qos_rules = self.api.qos_list(
+            self.switch.uuid).execute(check_error=True)
+        self.assertCountEqual([self.qos_3], qos_rules)
+
+    def test_qos_delete_external_ids_wrong_keys_or_values(self):
+        self._create_fip_qoses()
+        self.api.qos_del_ext_ids(self.switch.name,
+                                 {'key_z': 'value1'}).execute(check_error=True)
+        qos_rules = self.api.qos_list(
+            self.switch.uuid).execute(check_error=True)
+        self.assertCountEqual([self.qos_1, self.qos_2, self.qos_3], qos_rules)
+
+        self.api.qos_del_ext_ids(self.switch.name,
+                                 {'key1': 'value_z'}).execute(check_error=True)
+        qos_rules = self.api.qos_list(
+            self.switch.uuid).execute(check_error=True)
+        self.assertCountEqual([self.qos_1, self.qos_2, self.qos_3], qos_rules)
+
+        self.api.qos_del_ext_ids(
+            self.switch.name,
+            {'key3': 'value3', 'key4': 'value_z'}).execute(check_error=True)
+        qos_rules = self.api.qos_list(
+            self.switch.uuid).execute(check_error=True)
+        self.assertCountEqual([self.qos_1, self.qos_2, self.qos_3], qos_rules)
+
+    def test_qos_delete_external_ids_empty_dict(self):
+        self.assertRaises(TypeError, self.api.qos_del_ext_ids,
+                          self.switch.name, {})
+
+    def test_qos_delete_external_ids_if_exists(self):
+        self._create_fip_qoses()
+        cmd = self.api.qos_del_ext_ids('wrong_ls_name',
+                                       {'key1': 'value1'}, if_exists=False)
+        self.assertRaises(RuntimeError, cmd.execute, check_error=True)
+
+        self.api.qos_del_ext_ids('wrong_ls_name',
+                                 {'key1': 'value1'}).execute(check_error=True)
+        # No qos rule has been deleted from the correct logical switch.
+        qos_rules = self.api.qos_list(
+            self.switch.uuid).execute(check_error=True)
+        self.assertCountEqual([self.qos_1, self.qos_2, self.qos_3], qos_rules)
+
 
 class TestLspOps(OvnNorthboundTest):
     def setUp(self):
         super(TestLspOps, self).setUp()
         name = utils.get_rand_device_name()
         self.switch = self.useFixture(
-            fixtures.LogicalSwitchFixture(name)).obj
+            fixtures.LogicalSwitchFixture(self.api, name)).obj
 
     def _lsp_add(self, switch, name, *args, **kwargs):
         name = utils.get_rand_device_name() if name is None else name
@@ -356,7 +429,7 @@ class TestLspOps(OvnNorthboundTest):
         self.assertEqual(lsp1, lsp2)
 
     def test_lsp_add_may_exist_wrong_switch(self):
-        sw = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+        sw = self.useFixture(fixtures.LogicalSwitchFixture(self.api)).obj
         lsp = self._lsp_add(self.switch, None)
         self.assertRaises(RuntimeError, self._lsp_add, sw, lsp.name,
                           may_exist=True)
@@ -435,7 +508,7 @@ class TestLspOps(OvnNorthboundTest):
 
     def test_lsp_del_wrong_switch(self):
         lsp = self._lsp_add(self.switch, None)
-        sw = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+        sw = self.useFixture(fixtures.LogicalSwitchFixture(self.api)).obj
         cmd = self.api.lsp_del(lsp.uuid, sw.uuid)
         self.assertRaises(RuntimeError, cmd.execute, check_error=True)
 
@@ -460,7 +533,7 @@ class TestLspOps(OvnNorthboundTest):
     def test_lsp_list_no_switch(self):
         ports = {self._lsp_add(self.switch, None) for _ in range(3)}
         other_switch = self.useFixture(fixtures.LogicalSwitchFixture(
-            name=utils.get_rand_device_name())).obj
+            self.api, name=utils.get_rand_device_name())).obj
         other_port = self._lsp_add(other_switch, None)
         all_ports = set(self.api.lsp_list().execute(check_error=True))
         self.assertTrue((ports.union(set([other_port]))).issubset(all_ports))
@@ -489,7 +562,7 @@ class TestLspOps(OvnNorthboundTest):
     def test_lsp_set_addresses_invalid(self):
         self.assertRaises(
             TypeError,
-            self.api.lsp_set_addresses, 'fake', 'invalidaddress')
+            self.api.lsp_set_addresses, 'fake', ['invalidaddress'])
 
     def test_lsp_get_addresses(self):
         addresses = [
@@ -547,7 +620,7 @@ class TestLspOps(OvnNorthboundTest):
     def test_lsp_set_get_dhcpv4_options(self):
         lsp = self._lsp_add(self.switch, None)
         dhcpopt = self.useFixture(
-            fixtures.DhcpOptionsFixture('192.0.2.1/24')).obj
+            fixtures.DhcpOptionsFixture(self.api, '192.0.2.1/24')).obj
         self.api.lsp_set_dhcpv4_options(
             lsp.name, dhcpopt.uuid).execute(check_error=True)
         options = self.api.lsp_get_dhcpv4_options(
@@ -558,7 +631,7 @@ class TestLspOps(OvnNorthboundTest):
 class TestDhcpOptionsOps(OvnNorthboundTest):
     def _dhcpopt_add(self, cidr, *args, **kwargs):
         dhcpopt = self.useFixture(fixtures.DhcpOptionsFixture(
-            cidr, *args, **kwargs)).obj
+            self.api, cidr, *args, **kwargs)).obj
         self.assertEqual(cidr, dhcpopt.cidr)
         return dhcpopt
 
@@ -605,7 +678,7 @@ class TestDhcpOptionsOps(OvnNorthboundTest):
 class TestLogicalRouterOps(OvnNorthboundTest):
     def _lr_add(self, *args, **kwargs):
         lr = self.useFixture(
-            fixtures.LogicalRouterFixture(*args, **kwargs)).obj
+            fixtures.LogicalRouterFixture(self.api, *args, **kwargs)).obj
         self.assertIn(lr.uuid, self.api.tables['Logical_Router'].rows)
         return lr
 
@@ -733,7 +806,7 @@ class TestLogicalRouterOps(OvnNorthboundTest):
 
     def test_lr_nat_add_port(self):
         sw = self.useFixture(
-            fixtures.LogicalSwitchFixture()).obj
+            fixtures.LogicalSwitchFixture(self.api)).obj
         lsp = self.api.lsp_add(sw.uuid, utils.get_rand_device_name()).execute(
             check_error=True)
         lport, mac = (lsp.name, 'de:ad:be:ef:4d:ad')
@@ -761,7 +834,7 @@ class TestLogicalRouterOps(OvnNorthboundTest):
 
     def test_lr_nat_add_may_exist(self):
         sw = self.useFixture(
-            fixtures.LogicalSwitchFixture()).obj
+            fixtures.LogicalSwitchFixture(self.api)).obj
         lsp = self.api.lsp_add(sw.uuid, utils.get_rand_device_name()).execute(
             check_error=True)
         args = (const.NAT_BOTH, '10.17.4.1', '192.0.2.1')
@@ -776,7 +849,7 @@ class TestLogicalRouterOps(OvnNorthboundTest):
 
     def test_lr_nat_add_may_exist_remove_port(self):
         sw = self.useFixture(
-            fixtures.LogicalSwitchFixture()).obj
+            fixtures.LogicalSwitchFixture(self.api)).obj
         lsp = self.api.lsp_add(sw.uuid, utils.get_rand_device_name()).execute(
             check_error=True)
         args = (const.NAT_BOTH, '10.17.4.1', '192.0.2.1')
@@ -878,7 +951,7 @@ class TestLogicalRouterOps(OvnNorthboundTest):
 class TestLogicalRouterPortOps(OvnNorthboundTest):
     def setUp(self):
         super(TestLogicalRouterPortOps, self).setUp()
-        self.lr = self.useFixture(fixtures.LogicalRouterFixture()).obj
+        self.lr = self.useFixture(fixtures.LogicalRouterFixture(self.api)).obj
 
     def _lrp_add(self, port, mac='de:ad:be:ef:4d:ad',
                  networks=None, *args, **kwargs):
@@ -932,7 +1005,7 @@ class TestLogicalRouterPortOps(OvnNorthboundTest):
     def test_lrp_add_may_exist_different_router(self):
         name = utils.get_rand_device_name()
         args = (name, 'de:ad:be:ef:4d:ad', ['192.0.2.0/24'])
-        lr2 = self.useFixture(fixtures.LogicalRouterFixture()).obj
+        lr2 = self.useFixture(fixtures.LogicalRouterFixture(self.api)).obj
         self._lrp_add(*args)
         cmd = self.api.lrp_add(lr2.uuid, *args, may_exist=True)
         self.assertRaises(RuntimeError, cmd.execute, check_error=True)
@@ -998,7 +1071,7 @@ class TestLogicalRouterPortOps(OvnNorthboundTest):
 
     def test_lrp_del_wrong_router(self):
         lrp = self._lrp_add(None)
-        sw = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+        sw = self.useFixture(fixtures.LogicalSwitchFixture(self.api)).obj
         cmd = self.api.lrp_del(lrp.uuid, sw.uuid)
         self.assertRaises(RuntimeError, cmd.execute, check_error=True)
 
@@ -1046,7 +1119,7 @@ class TestLoadBalancerOps(OvnNorthboundTest):
     def _lb_add(self, lb, vip, ips, protocol=const.PROTO_TCP, may_exist=False,
                 **columns):
         lbal = self.useFixture(fixtures.LoadBalancerFixture(
-            lb, vip, ips, protocol, may_exist, **columns)).obj
+            self.api, lb, vip, ips, protocol, may_exist, **columns)).obj
         self.assertEqual(lb, lbal.name)
         norm_vip = ovsdb_utils.normalize_ip_port(vip)
         self.assertIn(norm_vip, lbal.vips)
@@ -1126,6 +1199,32 @@ class TestLoadBalancerOps(OvnNorthboundTest):
         self.api.lb_del(utils.get_rand_device_name(), if_exists=True).execute(
             check_error=True)
 
+    def test_lb_del_vip_if_exists(self):
+        name = utils.get_rand_device_name()
+        lb = self._lb_add(name, '192.0.2.1', ['10.0.0.1'])
+        lb2 = self._lb_add(name, '192.0.2.2', ['10.1.0.1'])
+        self.assertEqual(lb, lb2)
+        # Remove vip that does not exist.
+        self.api.lb_del(lb.name, '9.9.9.9', if_exists=True
+                        ).execute(check_error=True)
+        self.assertIn('192.0.2.1', lb.vips)
+        self.assertIn('192.0.2.2', lb.vips)
+        # Remove one vip.
+        self.api.lb_del(lb.name, '192.0.2.1', if_exists=True
+                        ).execute(check_error=True)
+        self.assertNotIn('192.0.2.1', lb.vips)
+        # Attempt to remove same vip a second time with if_exists=True.
+        self.api.lb_del(lb.name, '192.0.2.1', if_exists=True).execute(
+            check_error=True)
+        # Attempt to remove same vip a third time with if_exists=False.
+        cmd = self.api.lb_del(lb.name, '192.0.2.1', if_exists=False)
+        self.assertRaises(idlutils.RowNotFound, cmd.execute, check_error=True)
+        # Remove last vip and make sure that lb is also removed.
+        self.assertIn('192.0.2.2', lb.vips)
+        self.api.lb_del(lb.name, '192.0.2.2').execute(check_error=True)
+        cmd = self.api.lb_del(lb.name)
+        self.assertRaises(idlutils.RowNotFound, cmd.execute, check_error=True)
+
     def test_lb_list(self):
         lbs = {self._lb_add(utils.get_rand_device_name(), '192.0.2.1',
                             ['10.0.0.1', '10.0.0.2']) for _ in range(3)}
@@ -1151,13 +1250,13 @@ class TestObLbOps(testscenarios.TestWithScenarios, OvnNorthboundTest):
         # They must be in this order because the load balancer
         # can't be deleted when there is a reference in the router
         self.lb = self.useFixture(fixtures.LoadBalancerFixture(
-            utils.get_rand_device_name(), '192.0.2.1',
+            self.api, utils.get_rand_device_name(), '192.0.2.1',
             ['10.0.0.1', '10.0.0.2'])).obj
         self.lb2 = self.useFixture(fixtures.LoadBalancerFixture(
-            utils.get_rand_device_name(), '192.0.2.2',
+            self.api, utils.get_rand_device_name(), '192.0.2.2',
             ['10.1.0.1', '10.1.0.2'])).obj
         self.lr = self.useFixture(self.fixture(
-            utils.get_rand_device_name())).obj
+            self.api, utils.get_rand_device_name())).obj
 
     def test_ob_lb_add(self):
         self.add_fn(self.lr.name, self.lb.name).execute(
@@ -1214,7 +1313,8 @@ class TestCommonDbOps(OvnNorthboundTest):
     def setUp(self):
         super(TestCommonDbOps, self).setUp()
         name = utils.get_rand_device_name()
-        self.switch = self.useFixture(fixtures.LogicalSwitchFixture(name)).obj
+        self.switch = self.useFixture(
+            fixtures.LogicalSwitchFixture(self.api, name)).obj
         self.lsps = [
             self.api.lsp_add(
                 self.switch.uuid,
@@ -1272,7 +1372,8 @@ class TestCommonDbOps(OvnNorthboundTest):
 
 class TestDnsOps(OvnNorthboundTest):
     def _dns_add(self, *args, **kwargs):
-        dns = self.useFixture(fixtures.DnsFixture(*args, **kwargs)).obj
+        dns = self.useFixture(
+            fixtures.DnsFixture(self.api, *args, **kwargs)).obj
         return dns
 
     def test_dns_get(self):
@@ -1340,11 +1441,13 @@ class TestDnsOps(OvnNorthboundTest):
 
 class TestLsDnsOps(OvnNorthboundTest):
     def _dns_add(self, *args, **kwargs):
-        dns = self.useFixture(fixtures.DnsFixture(*args, **kwargs)).obj
+        dns = self.useFixture(
+            fixtures.DnsFixture(self.api, *args, **kwargs)).obj
         return dns
 
     def _ls_add(self, *args, **kwargs):
-        fix = self.useFixture(fixtures.LogicalSwitchFixture(*args, **kwargs))
+        fix = self.useFixture(
+            fixtures.LogicalSwitchFixture(self.api, *args, **kwargs))
         return fix.obj
 
     def test_ls_dns_set_clear_records(self):
@@ -1384,7 +1487,8 @@ class TestPortGroup(OvnNorthboundTest):
 
     def setUp(self):
         super(TestPortGroup, self).setUp()
-        self.switch = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+        self.switch = self.useFixture(
+            fixtures.LogicalSwitchFixture(self.api)).obj
         self.pg_name = 'testpg-%s' % ovsdb_utils.generate_uuid()
 
     def test_port_group(self):
@@ -1546,3 +1650,44 @@ class TestHAChassisGroup(OvnNorthboundTest):
         hcg1 = cmd.execute(check_error=True)
         hcg2 = cmd.execute(check_error=True)
         self.assertEqual(hcg1, hcg2)
+
+
+class TestReferencedObjects(OvnNorthboundTest):
+    """Exercise adding a ls, lsp and lsp_address in a single transaction.
+
+    The main goal of this test is to make sure a transaction can use either
+    a name or an object notation in order to create an ls+lsp while in a
+    transaction context.
+    """
+    def setUp(self):
+        super(TestReferencedObjects, self).setUp()
+        self.ls_name = utils.get_rand_device_name()
+        self.lsp_name = utils.get_rand_device_name()
+        self.lsp_test_addresses = ['de:ad:be:ef:4d:ad 192.0.2.1']
+
+    def _check_values(self):
+        # Check: Make sure ls_get and lsp_get work (no RowNotFound exception)
+        self.api.ls_get(self.ls_name).execute(check_error=True)
+        self.api.lsp_get(self.lsp_name).execute(check_error=True)
+        self.assertEqual(self.lsp_test_addresses,
+                         self.api.lsp_get_addresses(self.lsp_name).execute(
+                             check_error=True))
+
+    def test_lsp_add_by_name(self):
+        with self.api.transaction(check_error=True) as txn:
+            txn.add(self.api.ls_add(self.ls_name))
+            txn.add(self.api.lsp_add(self.ls_name, self.lsp_name))
+            txn.add(self.api.lsp_set_addresses(self.lsp_name,
+                                               self.lsp_test_addresses))
+        self._check_values()
+
+    def test_lsp_add_by_object_via_db_create(self):
+        with self.api.transaction(check_error=True) as txn:
+            sw = txn.add(self.api.db_create_row('Logical_Switch',
+                                                name=self.ls_name))
+            prt = txn.add(self.api.db_create_row('Logical_Switch_Port',
+                                                 name=self.lsp_name))
+            txn.add(self.api.db_add('Logical_Switch', sw, "ports", prt))
+            txn.add(self.api.db_add('Logical_Switch_Port', prt,
+                                    "addresses", self.lsp_test_addresses[0]))
+        self._check_values()
