@@ -14,7 +14,6 @@
 from __future__ import print_function
 import atexit
 import os
-import six
 import subprocess
 import sys
 
@@ -22,8 +21,21 @@ from fixtures import fixture
 
 from ovsdbapp import venv
 
-if len(sys.argv) != 3:
-    print("Requires two arguments: venvdir ovsdir", file=sys.stderr)
+
+def reraise(tp, value, tb=None):
+    try:
+        if value is None:
+            value = tp()
+        if value.__traceback__ is not tb:
+            raise value.with_traceback(tb)
+        raise value
+    finally:
+        value = None
+        tb = None
+
+
+if len(sys.argv) != 4:
+    print("Requires three arguments: venvdir ovsdir ovndir", file=sys.stderr)
     sys.exit(1)
 
 for d in sys.argv[1:]:
@@ -33,13 +45,14 @@ for d in sys.argv[1:]:
 
 venvdir = os.path.abspath(sys.argv[1])
 ovsdir = os.path.abspath(sys.argv[2])
+ovndir = os.path.abspath(sys.argv[3])
 
-v = venv.OvsOvnVenvFixture(venvdir, ovsdir)
+v = venv.OvsOvnVenvFixture(venvdir, ovsdir, ovndir=ovndir)
 try:
     atexit.register(v.cleanUp)
     v.setUp()
 except fixture.MultipleExceptions as e:
-    six.reraise(*e.args[0])
+    reraise(*e.args[0])
 try:
     print("*** Exit the shell when finished debugging ***")
     subprocess.call([os.getenv('SHELL'), '-i'], env=v.env)
